"""
DjangoDumpDie testing URL Configuration.

"""

# Third-Party Imports.
from django.urls import path

# Internal Imports.
from .views.test_views import (
    index,

    complex_type_example__set,
    complex_type_example__frozen_set,
    complex_type_example__tuple,
    complex_type_example__list,
    complex_type_example__dict,
    complex_type_example__querydict,
    complex_type_example__memory_view,
    complex_type_example__enum,

    complex_type_example__multilevel_set,
    complex_type_example__multilevel_tuple,
    complex_type_example__multilevel_list,
    complex_type_example__multilevel_dict,

    complex_type_example__list_subitem,
    complex_type_example__tuple_subitem,
    complex_type_example__tuple_subitem_func,
    complex_type_example__dict_subitem,
    complex_type_example__enum_subitem,
)


app_name = 'django_dump_die_tests'
urlpatterns = [
    # Various "complex type" example views.
    path('complex/set/', complex_type_example__set, name='complex__set'),
    path('complex/frozen_set/', complex_type_example__frozen_set, name='complex__frozen_set'),
    path('complex/tuple/', complex_type_example__tuple, name='complex__tuple'),
    path('complex/list/', complex_type_example__list, name='complex__list'),
    path('complex/dict/', complex_type_example__dict, name='complex__dict'),
    path('complex/querydict/', complex_type_example__querydict, name='complex__querydict'),
    path('complex/memory_view/', complex_type_example__memory_view, name='complex__memory_view'),
    path('complex/enum/', complex_type_example__enum, name='complex__enum'),

    path('complex/multi_level/set/', complex_type_example__multilevel_set, name='complex__multi_level__set'),
    path('complex/multi_level/tuple/', complex_type_example__multilevel_tuple, name='complex__multi_level__tuple'),
    path('complex/multi_level/list/', complex_type_example__multilevel_list, name='complex__multi_level__list'),
    path('complex/multi_level/dict/', complex_type_example__multilevel_dict, name='complex__multi_level__dict'),

    path('complex/sub_item/list/', complex_type_example__list_subitem, name='complex__sub_item__list'),
    path('complex/sub_item/tuple/', complex_type_example__tuple_subitem, name='complex__sub_item__tuple'),
    path(
        'complex/sub_item/tuple_func/',
        complex_type_example__tuple_subitem_func,
        name='complex__sub_item__tuple_func',
    ),
    path('complex/sub_item/dict/', complex_type_example__dict_subitem, name='complex__sub_item__dict'),
    path('complex/sub_item/enum/', complex_type_example__enum_subitem, name='complex__sub_item__enum'),

    path('', index, name='index'),
]



"""Utils for dump die"""

# System Imports.
import copy
import inspect
import io
import linecache
import re
import tokenize
from collections.abc import Sequence
from decimal import Decimal
from enum import EnumMeta, Enum as OrigEnum
from tokenize import (
    generate_tokens,
    ENDMARKER,
    NAME,
    NEWLINE,
    NUMBER,
    OP,
    STRING,
)

# Third-Party Imports.
from django.core.exceptions import ObjectDoesNotExist
from django.http import QueryDict

# Internal Imports.
from django_dump_die.constants import COLORIZE_DUMPED_OBJECT_NAME, INCLUDE_FILENAME_LINENUMBER, PYTZ_PRESENT

# Imports that may not be accessible, depending on local python environment setup.
if PYTZ_PRESENT:
    import pytz


class Enum:
    """Enum faker class so an entire Enum can be dumped correctly."""
    def __init__(self, *args, **kwargs):
        """Set attrs on this class for each kwarg passed in."""
        for key, val in kwargs.items():
            setattr(self, key, val)


def get_dumped_object_info(obj, index_range=None, deepcopy=False):
    """Get all the required information to be able to dump an object"""
    # Get object filename, linenumber, and name
    filename, linenumber, obj_name = get_dumped_object_name_and_location(obj)

    # Process the object name to add coloring and/or put in correct format for the template.
    obj_name = process_object_name(obj_name)

    # Handle if function.
    function_doc = None
    if callable(obj) and not inspect.isclass(obj):
        function_doc = inspect.getdoc(obj)

    # Sanitize and validate provided index values.
    start_index, end_index = sanitize_index_range(index_range)

    # Handle if Enum by converting to a standard class.
    # NOTE: Dumping an Enum without conversion results in a blank string being dumped.
    if isinstance(obj, EnumMeta):
        obj = Enum(**obj.__members__)

    # Handle if deepcopy set.
    original_obj = None
    try:
        if deepcopy:
            original_obj = obj
            obj = copy.deepcopy(obj)
    except TypeError as type_error:
        original_obj = None
        raise TypeError(f"Object contains type that can't be deep copied. - {type_error}") from None

    return (filename, linenumber, obj_name, obj, function_doc, start_index, end_index, original_obj)


def get_dumped_object_name_and_location(object_needing_name):
    """Look at the stack frame to figure out what the dumped var is"""

    # Get the frame where the dump or dd occurred.
    base_frame = inspect.currentframe().f_back.f_back.f_back
    frame_info = inspect.getframeinfo(base_frame)

    # Default to not showing filename and lineno via setting to None
    linenumber = None
    filename = None
    # If we are to show the filename and linenumber
    if INCLUDE_FILENAME_LINENUMBER:
        # Get line number
        linenumber = frame_info.lineno
        # Get filename
        filename = frame_info.filename

    # Parse line that called, for full "name" to output.
    # Accounts for things like functions that span multiple lines.
    code_context = get_fully_qualified_dumped_line(base_frame, frame_info)

    # TODO: This work of using RegEx to get down to the dumped object name can probably be improved.
    # Establish a couple of regular expressions to fetch out what was passed to dump or dd.
    dump_pattern = r".*dump\((.*)[\)]"
    dd_pattern = r".*dd\((.*)[\)]"
    options_pattern = r"(.*)(?=,.*=.*)"
    # Find the results for both dump and dd.
    dumped_text_matches = re.findall(dump_pattern, code_context)
    dd_text_matches = re.findall(dd_pattern, code_context)
    # Determine if dumped or dd'd and put result in dumped_text
    dumped_text = 'Unknown_Object_Name'
    if dumped_text_matches:
        dumped_text = dumped_text_matches[0]
    elif dd_text_matches:
        dumped_text = dd_text_matches[0]

    while 'deepcopy' in dumped_text or 'index_range' in dumped_text:
        dumped_text_matches = re.findall(options_pattern, dumped_text)
        if dumped_text_matches:
            dumped_text = dumped_text_matches[0]
        else:
            break

    # If function get the callable name for each function name.
    if callable(object_needing_name) and not inspect.isclass(object_needing_name):
        dumped_text = get_callable_name(dumped_text, object_needing_name)

    return filename, linenumber, dumped_text


def get_fully_qualified_dumped_line(base_frame, frame_info):
    """Loop through frame info until we get full object name.

    Required for definitions of things that span multiple lines, within a given dump/dd statement.
    """

    line_num = base_frame.f_lineno - 1
    code_context = ''
    counter = None
    total_loops = 0
    # Loop through lines until counter goes back to 0, to account for newlines.
    # Counter increments for every proper ( token found, and decrements for every proper ) token found.
    # Additional checks are present so that the page does not hang forever, in event of input we did not account for.
    while counter is None or (0 < counter < 100 and total_loops < 1000):
        total_loops += 1

        # Initialize counter for first loop.
        if counter is None:
            counter = 0

        # Check line for parsing actual name info.
        line_num += 1
        line = linecache.getline(frame_info.filename, line_num, base_frame.f_globals)
        code_context += line

        # Get the tokens out of the string.
        # NOTE: generate_tokens returns an iterator and not a data structure.
        tokens = generate_tokens(io.StringIO(line).readline)

        # Iterate through all tokens from line, and adjust counter if paren tokens are found.
        try:
            for token in tokens:
                if token.exact_type == 7:
                    counter += 1
                elif token.exact_type == 8:
                    counter -= 1

        except tokenize.TokenError:
            # Seems to only come up for end of line. We can safely ignore, probably?
            pass

    # Strip out extra whitespace, if present.
    code_context = re.sub(r'\s+', ' ', code_context)

    # Return final result.
    return code_context


def process_object_name(object_name):
    """Process the object name into a dictionary of parts that can be colorized in output"""

    # List to return for the name
    name = []

    # If the dumped object name should be colorized
    if COLORIZE_DUMPED_OBJECT_NAME:

        # List of function names
        functions = []

        # List of params
        params = []

        # Get the tokens out of the string
        # NOTE: generate_tokens returns an iterator and not a data structure.
        tokens = generate_tokens(io.StringIO(object_name).readline)

        # Stores all tokens extracted from the iterator.
        all_tokens = []
        # Stores the previous token as we iterate.
        previous_token = ''
        # Whether we are inside a param list. Default to False
        in_param_list = False

        # Convert the tokens from the iterator to a list of tokens.
        # Also, store the tokens that are function names or params for later color processing.
        for token in tokens:
            # Store token for later processing.
            all_tokens.append(token)
            # Convert for readability
            token_number = token[0]
            token_value = token[1]

            # If token is a opening paren
            if token_number == OP and token.exact_type == 7:
                # Append previous token to the function list
                functions.append(previous_token)
                # Flip flag to indicate that we are now processing the parameter list.
                in_param_list = True

            # If we are inside the param list and not at the ending paren
            if in_param_list and token.exact_type != 8:
                # Append the token to the list of params
                params.append(token_value)
            else:
                in_param_list = False

            # Update the previous token to the current one.
            previous_token = token_value

        # Iterate over tokens classifying them with a CSS class.
        for token_number, token_value, _, _, _ in all_tokens:

            # Ignore newlines and endmarkers
            if token_number == NEWLINE or token_number == ENDMARKER:
                continue

            if token_number == NAME and token_value == 'Unknown_Object_Name':
                # Use empty color as object name can't be determined
                css_class = 'empty'
            elif token_number == NAME and is_const(token_value):
                # Use constant color.
                css_class = 'constant'
            elif token_number == STRING:
                # Use string color.
                css_class = 'string'
            elif token_number == NUMBER:
                # Use number color.
                css_class = 'number'
            elif token_number == OP and token_value in '(){}[]':
                # Use braces color.
                css_class = 'braces'
            elif token_value in functions:
                # Use function color
                css_class = 'function'
            elif token_value in params:
                # User param color
                css_class = 'params'
            else:
                # Use default color.
                css_class = 'dumped_name'

            # Append the info on to the name list
            name.append({
                'css_class': css_class,
                'value': token_value,
            })

    else:
        # Convert object name to the format expected by the template.
        name.append({
            'css_class': '',
            'value': object_name,
        })

    return name


def sanitize_index_range(index_range):
    """
    Validates and sanitizes passed index values.
    """
    # Set index defaults.
    start_index = None
    end_index = None

    # First check if index_range is an iterable.
    if isinstance(index_range, Sequence):
        # Handle for length 2 or more (we ignore values past second index).
        if len(index_range) > 1:
            # Assume is index range. Split values onto indexes.
            start_index = index_range[0]
            end_index = index_range[1]

        # Handle for length exactly 1.
        elif len(index_range) == 1:
            # Assume user typo'd somehow and meant literal start index.
            start_index = index_range[0]

    else:
        # Non-iterable. Assume user passed single start-index value.
        start_index = index_range

    # Sanitize if start_index set.
    if start_index:
        try:
            start_index = int(start_index)
        except TypeError:
            start_index = None

    # Sanitize if end_index set.
    if end_index:
        try:
            end_index = int(end_index)
        except Exception:
            end_index = None

    return start_index, end_index


# region Object Property Functions


def generate_unique_from_obj(obj):
    """Generate a unique identifier for the object passed in."""

    # Create unique via hash and fallback to id on exception.
    try:
        unique = hash(obj)
    except Exception:
        unique = id(obj)
    # Append the class name to the unique to really make unique.
    unique = f'{get_class_name(obj)}_{unique}'

    return unique


def get_members(obj):
    """Attempts to get object members. Falls back to an empty list."""

    # Get initial member set or empty list.
    members = inspect.getmembers(obj)

    # Add type specific members that will not be included from the use of the inspect.getmembers function.
    if is_dict(obj):
        if isinstance(obj, QueryDict):
            # Django QueryDict members. Has handling for multiple unique values referencing the same key.
            # https://docs.djangoproject.com/en/dev/ref/request-response/#querydict-objects
            obj_copy = copy.deepcopy(obj)
            for key in obj.keys():
                item = obj_copy.pop(key)
                members.append((key, item))
        else:
            # Standard dictionary members.
            members.extend(obj.items())

    elif is_iterable(obj):
        # Lists, sets, etc.
        if _is_indexable(obj):
            # Use indexes as left half.
            members.extend(list(enumerate(obj)))
        elif is_set(obj):
            # Use None as left half. Most likely a set.
            members.extend([(None, x) for x in obj])
    elif is_enum_member(obj):
        # Prevent enum entry from recursively showing up in entries members.
        members = [m for m in members if not is_enum_member(m[1])]

    return members


def get_class_name(obj):
    """Get class name of an object."""
    name = None
    try:
        name = obj.__class__.__name__
    except Exception:
        pass
    return name


def get_callable_name(obj_name, obj):
    """Get callable name of an object"""

    # Get the method signature and fall back to simply appending
    # parentheses to the method name on exception.
    try:
        obj_name += safe_str(inspect.signature(obj))
    except Exception:
        obj_name += '()'
    return obj_name


def get_callable_params(obj):
    """Get callable params of an object"""

    # Get the method signature and fall back to simply returning
    # parentheses on exception.
    try:
        signature = safe_str(inspect.signature(obj))
    except Exception:
        signature = '()'

    signature_minus_parentheses = signature[1:-1]

    return signature_minus_parentheses


def get_obj_type(obj):
    """Determines the string representation of object's type."""

    # Get default type value.
    obj_type = type(obj).__name__

    # Special handling for certain types.
    if obj_type == 'NoneType':
        obj_type = 'null'
    elif PYTZ_PRESENT and isinstance(obj, pytz.BaseTzInfo):
        obj_type = 'pytz_timezone'

    return obj_type


def safe_repr(obj):
    """Call repr() and ignore ObjectDoesNotExist."""
    str_obj = ''
    try:
        str_obj = repr(obj)
    except ObjectDoesNotExist:
        # NOTE: A list of deleted db objects will cause repr(list) to fail.
        # So, we detect this and print out the __class__ of the contents of
        # the list.
        str_obj = f'<{obj.__class__} DELETED>'

    return str_obj


def safe_str(obj):
    """Call str() and ignore TypeErrors if str() doesn't return a string."""
    str_obj = ''
    try:
        str_obj = str(obj)
    except (TypeError, ObjectDoesNotExist):
        str_obj = safe_repr(obj)

    return str_obj


def is_iterable(obj):
    """Return True if object can be iterated."""
    try:
        iter(obj) and len(obj)
    except NotImplementedError:
        return False
    except TypeError:
        return False
    except ValueError:
        return False
    return True


def is_query(obj):
    """Return True if object is most likely a query."""
    if obj is not None:
        return _in_dir(obj, 'as_manager') and _in_dir(obj, 'all') and _in_dir(obj, 'filter')


def is_dict(obj):
    """Return True if object is most likely a dict."""
    if obj is not None:
        return _in_dir(obj, 'items') and _in_dir(obj, 'keys') and _in_dir(obj, 'values')


def is_set(obj):
    """Return True if object is most likely a dict."""
    return isinstance(obj, set) or isinstance(obj, frozenset)

def is_enum_member(obj):
    """Return True if object is most likely a enum"""
    return issubclass(type(obj), OrigEnum)

def is_const(obj):
    """Return True if object is most likely a constant."""
    if obj is not None:
        return isinstance(obj, str) and obj[0].isalpha() and obj.upper() == obj


def is_key(obj):
    """Return True if object is most likely a key."""
    if obj is not None and isinstance(obj, str):
        return "'" in obj


def is_number(obj):
    """Return True if object is most likely a number."""
    if obj is not None:
        return (
            isinstance(obj, (int, float, Decimal))
            or (getattr(obj, 'isnumeric', None) and obj.isnumeric())
        )


def is_private(obj):
    """Return True if object is private."""
    if obj is not None:
        return isinstance(obj, str) and obj.startswith('_')


def is_magic(obj):
    """Return True if object is private."""
    if obj is not None:
        return isinstance(obj, str) and obj.startswith('__') and obj.endswith('__')


def _in_dir(obj, attr):
    """Simpler hasattr() function without side effects."""
    return attr in dir(obj)


def _is_indexable(obj):
    """Return True if object can be indexed."""
    if isinstance(obj, Sequence):
        return True
    else:
        return False

# endregion Object Property Functions



"""Template Tags for DumpDie"""

# System Imports.
import inspect
import re
import traceback
import types
from decimal import Decimal

# Third-Party Imports.
from django import template
from django.forms.boundfield import BoundField

# Internal Imports.
from django_dump_die.constants import (
    SIMPLE_TYPES,
    INTERMEDIATE_TYPES,
    ADDITIONAL_SIMPLE_TYPES,
    ADDITIONAL_INTERMEDIATE_TYPES,
    MAX_RECURSION_DEPTH,
    MAX_ITERABLE_LENGTH,
    MULTILINE_FUNCTION_DOCS,
    CONTENT_STARTS_EXPANDED,
    ATTRIBUTES_START_EXPANDED,
    FUNCTIONS_START_EXPANDED,
    INCLUDE_PRIVATE_METHODS,
    INCLUDE_MAGIC_METHODS,
    INCLUDE_ATTRIBUTES,
    INCLUDE_FUNCTIONS,
    PYTZ_PRESENT,
)
from django_dump_die.utils import (
    get_dumped_object_info,
    generate_unique_from_obj,
    get_members,
    get_class_name,
    get_callable_params,
    get_obj_type,
    safe_repr,
    safe_str,
    is_iterable,
    is_query,
    is_dict,
    is_const,
    is_key,
    is_number,
    is_private,
    is_magic,
)

# Imports that may not be accessible, depending on local python environment setup.
if PYTZ_PRESENT:
    import pytz


register = template.Library()


# region Module Variables

# Stores the uniques for each dumped root object.
repeat_iteration_tracker = {}
deepcopy_unique_map = {}

# endregion Module Variables


@register.inclusion_tag('django_dump_die/partials/_dump.html', takes_context=True)
def dump(context, obj):
    """Template tag that can be used in templates to use dd"""
    object_info = get_dumped_object_info(obj)

    render_head = context.get('django_dd_template_tag_render_head', True)
    if render_head:
        context['django_dd_template_tag_render_head'] = False

    return {
        'objects': [object_info],
        'render_head': render_head,
    }


@register.inclusion_tag('django_dump_die/partials/_dump_objects.html')
def dump_objects(objects):
    """Template tag that can be used to dump a list of objects"""
    return {'objects': objects}


@register.inclusion_tag('django_dump_die/partials/_dump_object.html')
def dump_object(
    obj,
    root_obj,
    skip_set=None,
    current_iteration=0,
    current_depth=0,
    root_index_start=None,
    root_index_end=None,
    original_obj=None,
    parent_is_intermediate=False,
):
    """Determines template dd/dump info for provided object.

    Determines appropriate attributes/functions associated with object, for dd/dump output to template.

    There are three major object categories:
      * Simple - Basic entities that simply output minimal, direct value output.
      * Intermediate - Objects that that display only one level of attribute/function values.
      * Complex - Complex objects that recursively display attribute/function values for self and all children.

    Note:
      * If we have exceeded specified iteration count or depth, then defaults back to "simple" type output.
      * On each call, inner children are minimally processed, and fully processed later in a new call to templatetag.

    :param obj: Object to iterate over and attempt to parse information from.
    :param root_obj: Root, parent object associated with current object.
    :param skip_set: Set of already-processed objects. Used to skip re-processing identical non-objects.
    :param current_iteration: Current iteration-index. Used to track current index of object we're iterating through.
    :param current_depth: Current depth-index. Used to track how deep of child-members we're iterating through.
    :param root_index_start: Starting index for root iterable object. If None, uses default behavior.
    :param root_index_end: Ending index for root iterable object. If None, uses default behavior.
    :param original_obj: Original object, for handling uniques if dealing with deepcopy.
    :param parent_is_intermediate: Boolean indicating that parent is intermediate type. Do not recurse further.
    """
    # Set up set to store uniques to skip if not passed in.
    # Will be used to skip objects already done to prevent infinite loops.
    skip_set = skip_set or set()

    # Generate the unique
    unique, root_count = _generate_unique(obj, root_obj, original_obj)

    # Following section will determine what should get rendered out.
    intermediate_value = None

    # Handle if object is in skip set, aka already processed.
    if unique in skip_set:
        # Complex object found in skip set. Skip further handling of if clauses and go to end of function.
        # Intermediates get slightly extra handling for "simple" value output.
        if _is_intermediate_type(obj):
            intermediate_value = safe_str(obj)

    # Handle if obj is a simple type (Null/None, int, str, bool, and basic number types)
    # OR if direct parent is an intermediate (excluding pytz timezone objects).
    elif (
        _is_simple_type(obj)
        or (
            (PYTZ_PRESENT and parent_is_intermediate and not isinstance(obj, pytz.BaseTzInfo))
            or parent_is_intermediate
        )
    ):
        return _handle_simple_type(obj)

    # Handle if obj is an intermediate (date/time types).
    elif _is_intermediate_type(obj):
        return _handle_intermediate_type(
            obj,
            root_obj,
            unique,
            root_count,
            skip_set=skip_set,
            original_obj=original_obj
        )

    # Handle if element is iterable and we are at the root's element direct children (depth of 1),
    elif is_iterable(root_obj) and current_depth == 0:

        # Handle unique indexing logic for root element.
        root_index_start, root_index_end = _process_root_indices(
            root_index_start,
            root_index_end,
            len(root_obj)
        )

        # Handle if current index is between root_index values.
        # Otherwise fallback to "already processed" logic.
        if current_iteration >= root_index_start and current_iteration < root_index_end:

            # Handle for new "unique" object output.
            return _handle_complex_type(
                obj,
                root_obj,
                unique,
                root_count,
                skip_set=skip_set,
                current_iteration=current_iteration,
                current_depth=current_depth,
                original_obj=original_obj,
            )

    # Handle if not at root element and/or "root_index" values are not set.
    elif is_complex_type(current_depth, current_iteration):

        # Handle for new "unique" object output.
        return _handle_complex_type(
            obj,
            root_obj,
            unique,
            root_count,
            skip_set=skip_set,
            current_iteration=current_iteration,
            current_depth=current_depth,
            root_index_start=root_index_start,
            root_index_end=root_index_end,
            original_obj=original_obj,
        )

    # If we're here then the object has already been processed before,
    # or we have reached the max depth or number of iterations
    # or outside the bounds of the root indexes to process.
    # In any case, just return the type and unique of the object for output.
    return {
        'type': get_obj_type(obj),
        'unique': unique,
        'root_count': root_count,
        'intermediate': intermediate_value,
    }


# region Unique Mapping Functions

def _generate_unique(obj, root_obj, original_obj):
    """Generates object "unique" identifier.

    If using dump/dd on multiple instances of a root object, tracks such and appends iteration counters appropriately.

    If dealing with deepcopy, creates mapping to track the original values, so end user can track evolution of a single
    variable over time.

    :param obj: Current object to determine unique for.
    :param root_obj: Root object associated with current obj. Used for deepcopy mapping logic.
    :param original_obj: Original instance of object corresponding to current obj. Used for deepcopy mapping logic.
    """
    # Get a unique for the object.
    unique = generate_unique_from_obj(obj)

    # Get a unique for the object.
    root_unique = generate_unique_from_obj(root_obj)

    # Default root_count to blank string
    root_count_string = ''

    # If there is an original_obj, we may need to create the unique map so that
    # we can restore the original uniques to the deepcopied object.
    if original_obj:
        # Ensure root unique in root_unique_map.
        if root_unique not in deepcopy_unique_map:
            # Create the unique mapping of current deepcopy to original object.
            _create_unique_map(obj, root_unique, original_obj)

        # Skip simple types and intermediate types
        if unique in deepcopy_unique_map[root_unique]:
            # Do unique swap so that both unique and root unique are the
            # same value as the original value before deep copying.
            unique = deepcopy_unique_map[root_unique][unique]
            root_unique = deepcopy_unique_map[root_unique][root_unique]

    # If the root unique is already in repeat_iteration_tracker.
    if root_unique in repeat_iteration_tracker:
        # Unique found in tracker.
        # Determine appended "iteration tracker" value for root unique and all associated children.

        # If obj and root_obj are the same, increment the count.
        if obj == root_obj:
            # Get the current count out.
            root_count = repeat_iteration_tracker[root_unique]
            # Increment the count.
            repeat_iteration_tracker[root_unique] += 1

        # Else set the root count to the value - 1 as it is a child of the
        # root unique and would otherwise have the wrong value.
        else:
            # Get the current count out.
            root_count = repeat_iteration_tracker[root_unique] - 1

        # If the root count is greater than zero, use it.
        if root_count > 0:
            # Update root_count_string.
            root_count_string = f'_{root_count}'
        else:
            root_count_string = ''

    else:
        # Unique not found in tracker. Add the unique to the repeat_iteration_tracker.
        repeat_iteration_tracker[root_unique] = 1

    return unique, root_count_string


def _create_unique_map(obj, root_unique, original_obj):
    """Create an entry in the root_unique_map for this object."""

    # Create a dict for that unique map.
    deepcopy_unique_map[root_unique] = {}

    # Begin recursively adding entries to the unique map.
    _add_unique_map_entry(obj, original_obj, root_unique)


def _add_unique_map_entry(obj, original_obj, root_unique):
    """Add a unique entry to the unique entry map."""

    # Calculate the obj unique and the original obj unique.
    obj_unique = generate_unique_from_obj(obj)
    original_obj_unique = generate_unique_from_obj(original_obj)

    # Add the new unique to the root unique map.
    deepcopy_unique_map[root_unique][obj_unique] = original_obj_unique

    # Attempt to get member values of object and original object.
    members = get_members(obj)
    original_members = get_members(original_obj)

    # Loop through members and recursively determine unique mappings.
    # We do this because the state of an object might change over time, so child values (and mappings) might be
    # different. Parent might also be of complex type with children that are also complex types.
    for attr, value in members:
        # Skip simple types and children of intermediate types.
        if _is_simple_type(value) or _is_intermediate_type(obj):
            continue
        # Skip private members if not including them and functions.
        if (is_private(attr) and not INCLUDE_PRIVATE_METHODS) or callable(value):
            continue

        # Loop through orig members looking for a match.
        for orig_attr, orig_value in original_members:
            # Skip simple types and children of intermediate types.
            if _is_simple_type(value) or _is_intermediate_type(original_obj):
                continue
            # Skip private members if not including them and functions.
            if (is_private(attr) and not INCLUDE_PRIVATE_METHODS) or callable(value):
                continue

            # If the attrs match, make recursive call to add more entries to the unique map.
            if orig_attr == attr:
                _add_unique_map_entry(value, orig_value, root_unique)

# endregion Unique Mapping Functions


# region Type Handling Functions

def _is_simple_type(obj):
    """Return if the obj is a simple type."""
    return (
        obj is None
        or type(obj) in SIMPLE_TYPES
        or get_class_name(obj) in ADDITIONAL_SIMPLE_TYPES
    )


def _is_intermediate_type(obj):
    """Return if the obj is an intermediate type."""

    # Special handling for pytz timezone objects.
    if PYTZ_PRESENT and isinstance(obj, pytz.BaseTzInfo):
        return True

    # Handling for all other objects.
    return (
        type(obj) in INTERMEDIATE_TYPES
        or get_class_name(obj) in ADDITIONAL_INTERMEDIATE_TYPES
    )


def is_complex_type(current_depth, current_iteration):
    """Return if we should render the full object"""
    return (
        # Ensure all dump calls are processed.
        current_depth == 0
        # Check for any nested objects
        or (
            # Check if the max_recursion is set to None or we have not reached it yet.
            (
                MAX_RECURSION_DEPTH is None
                or current_depth < MAX_RECURSION_DEPTH
            )

            # And if the max_iterable_length is set to None,
            # or we have not reached it yet or we are at the root level.
            and (
                MAX_ITERABLE_LENGTH is None
                or current_iteration < MAX_ITERABLE_LENGTH
            )
        )
    )


def _handle_simple_type(obj):
    """
    Logic for outputting information about a "simple type".
    Includes str, numbers, bools, etc.
    """
    # Determine which css class to use.
    css_class = ''
    if obj is None:
        css_class = 'none'
    elif isinstance(obj, str):
        css_class = 'string'
    elif isinstance(obj, bool):
        css_class = 'bool'
    elif isinstance(obj, (int, Decimal, float, bytes)):
        css_class = 'number'
    elif _is_intermediate_type(obj):
        css_class = 'intermediate'
    elif isinstance(obj, types.ModuleType):
        css_class = 'module'
    elif isinstance(obj, BoundField):
        css_class = 'bound'
    else:
        css_class = 'default'

    # Determine which output to use.
    if _is_intermediate_type(obj) or isinstance(obj, Decimal):
        output_value = safe_str(obj)
    else:
        output_value = safe_repr(obj)

    # Since simple type, return safe representation of simple type and
    # which css class to use.
    return {
        'simple': output_value,
        'type': get_obj_type(obj),
        'css_class': css_class,
    }


def _handle_intermediate_type(obj, root_obj, unique, root_count, skip_set=None, original_obj=None):
    """Handling for a intermediate object.

    Effectively, it's an object that's complex enough to have associated attributes and functions that we want to
    display in dd output. But for basic cases, treating it more almost as a "simple type" is generally enough
    information. Thus, it ends up being a combination of the two output formats.

    Similar to "simple type" in that we display base value output without needing to expand.

    Similar to "complex type" in that we display associated attributes, but do not fully recursively expand all.

    :param obj: Object to iterate over and attempt to parse information from.
    :param root_obj: Root, parent object associated with current object.
    :param unique: Unique identifier associated with current object.
    :param root_count: Count of how many times the root object has been dumped as a string prefixed with '_'.
    :param original_obj: Original object, for handling uniques if dealing with deepcopy.
    """
    # Add unique to skip so it won't be processed a second time by additional
    # recursive calls to this template tag.
    skip_set.add(unique)

    # Attempt to get corresponding attribute/function values of object.
    attributes, functions = get_obj_values(obj)

    # Return information required to render object.
    context = {
        'include_attributes': INCLUDE_ATTRIBUTES,
        'include_functions': INCLUDE_FUNCTIONS,
        'collapsable': _get_collapsable_values(),
        'braces': '{}',
        'object': obj,
        'intermediate': safe_str(obj),
        'unique': unique,
        'root_count': root_count,
        'type': get_obj_type(obj),
        'is_iterable': False,
        'depth': 0,
        'root_index_start': None,
        'root_index_end': None,
        'original_obj': original_obj,
    }
    if INCLUDE_ATTRIBUTES:
        context['attributes'] = attributes
        context['index'] = 0
        context['root_obj'] = root_obj
        context['skip'] = set()
    if INCLUDE_FUNCTIONS:
        context['functions'] = functions
        context['multiline_function_docs'] = MULTILINE_FUNCTION_DOCS

    return context


def _handle_complex_type(
    obj,
    root_obj,
    unique,
    root_count,
    skip_set=None,
    current_iteration=0,
    current_depth=0,
    root_index_start=None,
    root_index_end=None,
    original_obj=None,
):
    """
    Main logic for outputting information for a given "unique".

    :param obj: Object to iterate over and attempt to parse information from.
    :param root_obj: Root, parent object associated with current object.
    :param unique: Unique identifier associated with current object.
    :param root_count: Count of how many times the root object has been dumped as a string prefixed with '_'.
    :param original_obj: Original object, for handling uniques if dealing with deepcopy.
    :param skip_set: Set of already-processed objects. Used to skip re-processing identical objects.
    :param current_iteration: Current iteration-index. Used to track current index of object we're iterating through.
    :param current_depth: Current depth-index. Used to track how deep of child-members we're iterating through.
    :param root_index_start:
    :param root_index_end:
    """
    # Add unique to skip so it won't be processed a second time by additional
    # recursive calls to this template tag.
    skip_set.add(unique)

    # If the object is a query, evaluate it.
    # This prevents a crash because a lazy queryset has too many members.
    if is_query(obj):
        obj = list(obj)

    # Determine which type of braces should be used.
    if isinstance(obj, list):
        braces = '[]'
    elif isinstance(obj, tuple):
        braces = '()'
    else:
        braces = '{}'

    # Attempt to get corresponding attribute/function values of object.
    attributes, functions = get_obj_values(obj)

    is_iterable_obj = is_iterable(obj) and not is_dict(obj) and not isinstance(obj, memoryview)
    is_dict_obj = is_dict(obj)

    # Return information required to render object.
    context = {
        'include_attributes': INCLUDE_ATTRIBUTES,
        'include_functions': INCLUDE_FUNCTIONS,
        'collapsable': _get_collapsable_values(),
        'braces': braces,
        'object': obj,
        'unique': unique,
        'root_count': root_count,
        'type': get_obj_type(obj),
        'is_iterable': is_iterable_obj,
        'is_dict': is_dict_obj,
        'depth': current_depth,
        'root_index_start': root_index_start,
        'root_index_end': root_index_end,
        'original_obj': original_obj,
    }
    if INCLUDE_ATTRIBUTES:
        context['attributes'] = attributes
        context['index'] = current_iteration
        context['root_obj'] = root_obj
        context['skip'] = skip_set
    if INCLUDE_FUNCTIONS:
        context['functions'] = functions
        context['multiline_function_docs'] = MULTILINE_FUNCTION_DOCS

    return context

# endregion Type Handling Functions


# region Object Property Functions

def get_obj_values(obj):
    """Determines full corresponding member values (attributes/functions) of object."""

    # Initialize attribute/function lists. This is what we ultimately return.
    attributes = []     # (attr, value, access_modifier, css_class, title)
    functions = []      # (attr, doc, access_modifier)

    try:
        # Attempt to get member values of object. Falls back to empty list on failure.
        members = get_members(obj)

    except Exception as exception:
        # On exception, add exception to attributes and return the attributes.

        # First get exception data.
        tb = exception.__traceback__
        title = 'Exception Occurred\n\n'
        title += f'Exception Type: {type(exception).__name__}\n'
        title += f'Exception Value: {str(exception)}\n\n'
        title += 'Traceback (most recent call last):\n\n'
        for entry in traceback.format_tb(tb):
            title += f'{entry}\n'
        title += 'End of traceback'

        # Add exception data to attributes and return.
        attributes.append(['EXCEPTION', str(exception), None, 'empty', title])
        return (attributes, functions)

    # Once all members have been collected, attempt to figure out what type, access modifier, css class, and title
    # should be used for each attribute/function/value in the members.
    for attr, value in members:

        try:

            # Skip private members if not including them.
            if is_private(attr) and not INCLUDE_PRIVATE_METHODS:
                continue

            # Determine if the value is callable (function).
            # If so, Functions will just return documentation.
            is_callable = callable(value)
            if is_callable:  # Handle member functions.

                # Skip dunder (magic) methods if not including them.
                if is_magic(attr) and not INCLUDE_MAGIC_METHODS:
                    continue

                # Get the method signature for attr.
                params = get_callable_params(value)

                # Get the documentation for the method.
                value = inspect.getdoc(value)

                # Get the access modifier for the method.
                access_modifier = _get_access_modifier(attr)

                # Append function to list of functions
                functions.append([attr, params, value, access_modifier])

            else:  # Handle member attributes.

                # Always skip dunder attributes.
                if is_magic(attr):
                    continue

                # If attr is not None (anything but set) change to safe_repr.
                if attr is not None:
                    attr = safe_repr(attr)

                # If not a dict and not None, remove the outside quotes so
                # it looks more like an attribute and not a string.
                if not is_dict(obj) and attr is not None:
                    attr = re.sub("'", "", attr)

                # Determine what type attribute is so that the access modifier,
                # css class, and title can be appropriately set.
                # Processing order is: Index, const, key, set, attribute
                if is_number(attr):  # Index.
                    access_modifier = None
                    css_class = 'index'
                    title = 'Index'
                elif is_const(attr):  # Constant.
                    access_modifier = _get_access_modifier(attr)
                    css_class = 'constant'
                    title = 'Constant'
                elif is_key(attr):  # Key.
                    access_modifier = None
                    css_class = 'key'
                    title = 'Key'
                elif attr is None:  # Set.
                    access_modifier = None
                    css_class = ''
                    title = ''
                else:  # Class Attribute.
                    access_modifier = _get_access_modifier(attr)
                    css_class = 'attribute'
                    title = 'Attribute'

                # Append the attribute information to the list of attributes.
                attributes.append([attr, value, access_modifier, css_class, title])

        except Exception as inner_exception:
            # First get exception data.
            tb = inner_exception.__traceback__
            title = 'Exception:\n'
            while tb is not None:
                line_num = tb.tb_lineno
                name = tb.tb_frame.f_code.co_name
                filename = tb.tb_frame.f_code.co_filename
                tb = tb.tb_next
                title += '\n[ {0} : {1} ] {2}\n'.format(line_num, name, filename)

            # Add exception data to attributes and return.
            attributes.append(['EXCEPTION', str(inner_exception), None, 'empty', title])

    # Attempt to sort the functions and just ignore any errors.
    try:
        functions = sorted(functions)
    except Exception:
        pass  # Ignore sort errors.

    return (attributes, functions)


def _get_access_modifier(obj):
    """Return the access modifier that should be used."""
    if is_magic(obj):
        return '-'
    elif is_private(obj):
        return '#'
    else:
        return '+'

# endregion Object Property Functions


# region Misc Functions

def _process_root_indices(start, end, parent_length):
    """Parse and validate indexes into expected format. Allows use of user-specified index ranges on root element."""

    # Save for later processing
    orig_end = end

    # Handle defaults.
    # If we got this far, at least one is set. Make sure the other is set as well.
    if start is None:
        start = 0
    if end is None:
        end = MAX_ITERABLE_LENGTH

    # Handle if provided start_index is negative.
    if start < 0:
        start = parent_length + start

        # Reset if still negative.
        if start < 0:
            start = 0

    # If the original value of end is None, there is no specified end and
    # it makes sense to then run from the start, now that it is calculated,
    # to the max iterable length.
    if orig_end is None:
        end = start + MAX_ITERABLE_LENGTH

    # Handle if provided end_index is negative.
    if end < 0:
        end = parent_length + end

        # Reset if still negative.
        if end < 0:
            end = 0

    # Handle if user provided a start_index that is higher than end_index.
    if start > end:
        temp = start
        start = end
        end = temp

    # Return the processed indices.
    return start, end


def _get_collapsable_values():
    """Get the arrow and collapsable values"""

    # Determine default sets.
    content_set = {
        'arrow': '▼' if CONTENT_STARTS_EXPANDED else '▶',
        'show': 'show' if CONTENT_STARTS_EXPANDED else '',
        'always_show': 'false',
        'class': '' if CONTENT_STARTS_EXPANDED else 'collapsed',
        'aria': 'true' if CONTENT_STARTS_EXPANDED else 'false',
    }
    attr_set = {
        'arrow': '▼' if ATTRIBUTES_START_EXPANDED else '▶',
        'show': 'show' if ATTRIBUTES_START_EXPANDED else '',
        'always_show': 'false',
        'class': '' if ATTRIBUTES_START_EXPANDED else 'collapsed',
        'aria': 'true' if ATTRIBUTES_START_EXPANDED else 'false',
    }
    func_set = {
        'arrow': '▼' if FUNCTIONS_START_EXPANDED else '▶',
        'show': 'show' if FUNCTIONS_START_EXPANDED else '',
        'always_show': 'false',
        'class': '' if FUNCTIONS_START_EXPANDED else 'collapsed',
        'aria': 'true' if FUNCTIONS_START_EXPANDED else 'false',
    }

    # Extra handling if either attr or func output is disabled.
    # In such a case, it doesn't make sense to have expandable arrows for the remaining one.
    if INCLUDE_FUNCTIONS is False:
        attr_set['arrow'] = ''
        attr_set['show'] = 'show'
        attr_set['always_show'] = 'true'
        attr_set['class'] = 'show always-show'
        attr_set['aria'] = ''
    if INCLUDE_ATTRIBUTES is False:
        func_set['arrow'] = ''
        func_set['show'] = 'show'
        func_set['always_show'] = 'true'
        func_set['class'] = 'show always-show'
        func_set['aria'] = ''

    return {
        'content': content_set,
        'attribute': attr_set,
        'function': func_set,
    }

# endregion Misc Functions



"""
Tests for "intermediate" type DD output.
"""

# System Imports.
import datetime
import os
import unittest
from unittest.mock import patch

# Third-Party Imports.
from django.test import override_settings
from django.utils import timezone
from freezegun import freeze_time
from django_expanded_test_cases import IntegrationTestCase

# Internal Imports.
from django_dump_die.constants import PYTZ_PRESENT, ZONEINFO_PRESENT


# Set datetime for freezegun and template render comparison.
# Ensures that seconds, etc match exactly for what is rendered vs what we check.
# Note that we replace microseconds to ensure these values are NOT identical.
now_dt_time = datetime.datetime.now()
now_tz_time = timezone.now()
# now_dt_time = datetime.datetime.now().replace(microsecond=123456)
# now_tz_time = timezone.now().replace(microsecond=234567)
# now_dt_time = now_dt_time.replace(microsecond=123456)
# now_tz_time = now_tz_time.replace(microsecond=234567)

project_path = os.getcwd()


@override_settings(DEBUG=True)
class DumpDieIntermediateTypeTestCase(IntegrationTestCase):
    """Verify handling of dumped "intermediate" types."""
    url = 'django_dump_die:intermediate-type-example'

    def test_toolbar_display(self):
        """Verify page properly displays toolbar."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                # Check toolbar header.
                '<div class="dump-toolbar">',
                '<div><h1>Django DumpDie</h1></div>',

                # Check existence of buttons.
                '<p id="expand-all" class="button">Expand All</p>',
                '<p id="expand-1st-lvl" class="button">Expand First Level</p>',
                '<p id="expand-2nd-lvl" class="button">Expand Second Level</p>',
                '<p id="collapse-all" class="button">Collapse All</p>',
                '<p id="collapse-1st-lvl" class="button">Collapse First Level</p>',
                '<p id="collapse-2nd-lvl" class="button">Collapse Second Level</p>',

                '<div class="static-padding"></div>',
            ],
            content_starts_after='<body>',
            content_ends_before='<div class="dump-wrapper">',
        )

    def test_page_descriptor_display(self):
        """Test initial page descriptor output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                # Check page descriptor.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">'Displaying example of "intermediate type" object output.'</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'Displaying example of "intermediate type" object output.'</code>
                </div>
                """,
                '<hr>',
                # Check visual-padding lines.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">''</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </div>
                """,
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">''</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='<div class="static-padding"></div>',
            content_ends_before='Python type examples:',
        )

    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_bytes_array_display(self, mocked_unique_generation):
        """Verify dumping a "bytes array" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_bytes_array</span>
                    </span>:
                    <span class="type" title="bytearray">bytearray</span>
                    <code class="intermediate">bytearray(b'\\x08\\t\\n\\x0b')</code>
                    <span class="braces">{</span>
                    <a
                        class="arrow-toggle collapsed"
                        title="[Ctrl+click] Expand all children"
                        data-toggle="collapse"
                        data-target=".data_9001"
                        data-dd-type="type"
                        data-object-depth="1"
                        aria-label="Close"
                        aria-expanded="false"
                    >
                        <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                        <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                    </a>
                    <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                        <ul class="attribute-list">
                            <a
                                class="arrow-toggle show always-show"
                                title="[Ctrl+click] Expand all children"
                                data-target=".data_9001-attributes"
                                data-dd-type="attr"
                                aria-label="Open/Close"
                                aria-expanded=""
                            >
                                <span class="section_name">Attributes</span>
                                    <span id="arrow-data_9001-attributes" class="arrow">
                                </span>
                            </a>
                            <div
                                class="li-wrapper collapse data_9001-attributes show"
                                data-unique-attributes="data_9001-attributes"
                            >
                """,

                # Object child elements.
                """
                <li>
                    <span class="index" title="Index">0</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">8</code>
                </li>
                """,
                """
                <li>
                    <span class="index" title="Index">1</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">9</code>
                </li>
                """,
                """
                <li>
                    <span class="index" title="Index">2</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">10</code>
                </li>
                """,
                """
                <li>
                    <span class="index" title="Index">3</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">11</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='Python type examples:',
            content_ends_before='sample_complex',
        )

    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_complex_display(self, mocked_unique_generation):
        """Verify dumping a "complex" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_complex</span>
                    </span>:
                    <span class="type" title="complex">complex</span>
                    <code class="intermediate">(3-1j)</code>
                    <span class="braces">{</span>
                    <a
                        class="arrow-toggle collapsed"
                        title="[Ctrl+click] Expand all children"
                        data-toggle="collapse"
                        data-target=".data_9001"
                        data-dd-type="type"
                        data-object-depth="1"
                        aria-label="Close"
                        aria-expanded="false"
                    >
                        <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                        <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                    </a>
                    <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                        <ul class="attribute-list">
                            <a
                                class="arrow-toggle show always-show"
                                title="[Ctrl+click] Expand all children"
                                data-target=".data_9001-attributes"
                                data-dd-type="attr"
                                aria-label="Open/Close"
                                aria-expanded=""
                            >
                                <span class="section_name">Attributes</span>
                                    <span id="arrow-data_9001-attributes" class="arrow">
                                </span>
                            </a>
                            <div
                                class="li-wrapper collapse data_9001-attributes show"
                                data-unique-attributes="data_9001-attributes"
                            >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">imag</span>:
                    <span class="type" title="float">float</span>
                    <code class="number">-1.0</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">real</span>:
                    <span class="type" title="float">float</span>
                    <code class="number">3.0</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_bytes_array',
            content_ends_before='Date/Time examples:',
        )

    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_dt_date_display(self, mocked_unique_generation):
        """Verify dumping a "datetime date" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_dt_date</span>
                    </span>:
                    <span class="type" title="date">date</span>
                """
                """
                <code class="intermediate">{0}</code>
                """.format(now_dt_time.date()),
                """
                <span class="braces">{</span>
                <a
                    class="arrow-toggle collapsed"
                    title="[Ctrl+click] Expand all children"
                    data-toggle="collapse"
                    data-target=".data_9001"
                    data-dd-type="type"
                    data-object-depth="1"
                    aria-label="Close"
                    aria-expanded="false"
                >
                    <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                    <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                </a>
                <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                    <ul class="attribute-list">
                        <a
                            class="arrow-toggle show always-show"
                            title="[Ctrl+click] Expand all children"
                            data-target=".data_9001-attributes"
                            data-dd-type="attr"
                            aria-label="Open/Close"
                            aria-expanded=""
                        >
                            <span class="section_name">Attributes</span>
                                <span id="arrow-data_9001-attributes" class="arrow">
                            </span>
                        </a>
                """,

                # Object child elements.
                """
                <div
                    class="li-wrapper collapse data_9001-attributes show"
                    data-unique-attributes="data_9001-attributes"
                >
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">day</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.day),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">max</span>:
                    <span class="type" title="date">date</span>
                    <code class="intermediate">9999-12-31</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">min</span>:
                    <span class="type" title="date">date</span>
                    <code class="intermediate">0001-01-01</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">month</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.month),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">resolution</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">1 day, 0:00:00</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">year</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.year),

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='Date/Time examples:',
            content_ends_before='sample_tz_date',
        )

    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_tz_date_display(self, mocked_unique_generation):
        """Verify dumping a "timezone date" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_tz_date</span>
                    </span>:
                    <span class="type" title="date">date</span>
                """
                """
                <code class="intermediate">{0}</code>
                """.format(now_dt_time.date()),
                """
                <span class="braces">{</span>
                <a
                    class="arrow-toggle collapsed"
                    title="[Ctrl+click] Expand all children"
                    data-toggle="collapse"
                    data-target=".data_9001"
                    data-dd-type="type"
                    data-object-depth="1"
                    aria-label="Close"
                    aria-expanded="false"
                >
                    <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                    <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                </a>
                <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                    <ul class="attribute-list">
                        <a
                            class="arrow-toggle show always-show"
                            title="[Ctrl+click] Expand all children"
                            data-target=".data_9001-attributes"
                            data-dd-type="attr"
                            aria-label="Open/Close"
                            aria-expanded=""
                        >
                            <span class="section_name">Attributes</span>
                                <span id="arrow-data_9001-attributes" class="arrow">
                            </span>
                        </a>
                        <div
                            class="li-wrapper collapse data_9001-attributes show"
                            data-unique-attributes="data_9001-attributes"
                        >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">day</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.day),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">max</span>:
                    <span class="type" title="date">date</span>
                    <code class="intermediate">9999-12-31</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">min</span>:
                    <span class="type" title="date">date</span>
                    <code class="intermediate">0001-01-01</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">month</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.month),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">resolution</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">1 day, 0:00:00</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">year</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.year),

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_dt_date',
            content_ends_before='sample_dt_datetime',
        )

    # @freeze_time(now_dt_time)
    # @patch('django_dump_die.templatetags.dump_die._generate_unique')
    # def test_dt_datetime_display(self, mocked_unique_generation):
    #     """Verify dumping a "datetime datetime" type has expected output."""
    #
    #     # Override default "unique" generation logic, for reproduce-able tests.
    #     # This generates enough uniques to guarantee mock does not raise errors.
    #     side_effects = []
    #     for index in range(5000):
    #         side_effects += [
    #             (f'data_9001', ''),
    #         ]
    #     mocked_unique_generation.side_effect = side_effects
    #
    #     self.assertGetResponse(
    #         self.url,
    #         expected_title='DD',
    #         expected_header='Django DumpDie',
    #         expected_content=[
    #             '<hr>',
    #
    #             # Object opening tags.
    #             """
    #             <div class="dump-wrapper">
    #                 <span class="dumped_object" title="Dumped Object">
    #                     <span class="dumped_name">sample_dt_datetime</span>
    #                 </span>:
    #                 <span class="type" title="FakeDatetime">FakeDatetime</span>
    #             """
    #             """
    #             <code class="intermediate">{0}</code>
    #             """.format(now_dt_time),
    #             """
    #             <span class="braces">{</span>
    #             <a
    #                 class="arrow-toggle collapsed"
    #                 title="[Ctrl+click] Expand all children"
    #                 data-toggle="collapse"
    #                 data-target=".data_9001"
    #                 data-dd-type="type"
    #                 data-object-depth="1"
    #                 aria-label="Close"
    #                 aria-expanded="false"
    #             >
    #                 <span class="unique" data-highlight-unique="data_9001">data_9001</span>
    #                 <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
    #             </a>
    #             <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
    #                 <ul class="attribute-list">
    #                     <a
    #                         class="arrow-toggle show always-show"
    #                         title="[Ctrl+click] Expand all children"
    #                         data-target=".data_9001-attributes"
    #                         data-dd-type="attr"
    #                         aria-label="Open/Close"
    #                         aria-expanded=""
    #                     >
    #                         <span class="section_name">Attributes</span>
    #                             <span id="arrow-data_9001-attributes" class="arrow">
    #                         </span>
    #                     </a>
    #                     <div
    #                         class="li-wrapper collapse data_9001-attributes show"
    #                         data-unique-attributes="data_9001-attributes"
    #                     >
    #             """,
    #
    #             # Object child elements.
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">day</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_dt_time.day),
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">fold</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">0</code>
    #             </li>
    #             """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">hour</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_dt_time.hour),
    #             # """
    #             # <li>
    #             #     <span class="access-modifier">+</span>
    #             #     <span class="attribute" title="Attribute">max</span>:
    #             #     <span class="type" title="datetime">datetime</span>
    #             #     <code class="intermediate">9999-12-31 23:59:59.999999</code>
    #             # </li>
    #             # """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">microsecond</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_dt_time.microsecond),
    #             # """
    #             # <li>
    #             #     <span class="access-modifier">+</span>
    #             #     <span class="attribute" title="Attribute">min</span>:
    #             #     <span class="type" title="datetime">datetime</span>
    #             #     <code class="intermediate">0001-01-01 00:00:00</code>
    #             # </li>
    #             # """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">minute</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_dt_time.minute),
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">month</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_dt_time.month),
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">resolution</span>:
    #                 <span class="type" title="timedelta">timedelta</span>
    #                 <code class="intermediate">0:00:00.000001</code>
    #             </li>
    #             """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">second</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_dt_time.second),
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">tzinfo</span>:
    #                 <span class="type" title="null">null</span>
    #                 <code class="none">None</code>
    #             </li>
    #             """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">year</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_dt_time.year),
    #
    #             # Object closing tags.
    #             """
    #                                 </div>
    #                             </ul>
    #                         <ul class="attribute-list"></ul>
    #                     </div>
    #                 <span class="braces">}</span>
    #             </div>
    #             """,
    #
    #             '<hr>',
    #         ],
    #         content_starts_after='sample_tz_date',
    #         content_ends_before='sample_tz_datetime',
    #     )

    # @freeze_time(now_tz_time)
    # @patch('django_dump_die.templatetags.dump_die._generate_unique')
    # def test_tz_datetime_display(self, mocked_unique_generation):
    #     """Verify dumping a "timezone datetime" type has expected output."""
    #
    #     # Override default "unique" generation logic, for reproduce-able tests.
    #     # This generates enough uniques to guarantee mock does not raise errors.
    #     side_effects = []
    #     for index in range(5000):
    #         side_effects += [
    #             (f'data_9001', ''),
    #         ]
    #     mocked_unique_generation.side_effect = side_effects
    #
    #     self.assertGetResponse(
    #         self.url,
    #         expected_title='DD',
    #         expected_header='Django DumpDie',
    #         expected_content=[
    #             '<hr>',
    #
    #             # Object opening tags.
    #             """
    #             <div class="dump-wrapper">
    #                 <span class="dumped_object" title="Dumped Object">
    #                     <span class="dumped_name">sample_tz_datetime</span>
    #                 </span>:
    #                 <span class="type" title="FakeDatetime">FakeDatetime</span>
    #             """,
    #             """
    #             <code class="intermediate">{0}</code>
    #             """.format(now_tz_time),
    #             """
    #             <span class="braces">{</span>
    #             <a
    #                 class="arrow-toggle collapsed"
    #                 title="[Ctrl+click] Expand all children"
    #                 data-toggle="collapse"
    #                 data-target=".data_9001"
    #                 data-dd-type="type"
    #                 data-object-depth="1"
    #                 aria-label="Close"
    #                 aria-expanded="false"
    #             >
    #                 <span class="unique" data-highlight-unique="data_9001">data_9001</span>
    #                 <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
    #             </a>
    #             <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
    #                 <ul class="attribute-list">
    #                     <a
    #                         class="arrow-toggle show always-show"
    #                         title="[Ctrl+click] Expand all children"
    #                         data-target=".data_9001-attributes"
    #                         data-dd-type="attr"
    #                         aria-label="Open/Close"
    #                         aria-expanded=""
    #                     >
    #                         <span class="section_name">Attributes</span>
    #                             <span id="arrow-data_9001-attributes" class="arrow">
    #                         </span>
    #                     </a>
    #                     <div
    #                         class="li-wrapper collapse data_9001-attributes show"
    #                         data-unique-attributes="data_9001-attributes"
    #                     >
    #             """,
    #
    #             # Object child elements.
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">day</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_tz_time.day),
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">fold</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">0</code>
    #             </li>
    #             """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">hour</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_tz_time.hour),
    #             # """
    #             # <li>
    #             #     <span class="access-modifier">+</span>
    #             #     <span class="attribute" title="Attribute">max</span>:
    #             #     <span class="type" title="datetime">datetime</span>
    #             #     <code class="intermediate">9999-12-31 23:59:59.999999</code>
    #             # </li>
    #             # """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">microsecond</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_tz_time.microsecond),
    #             # """
    #             # <li>
    #             #     <span class="access-modifier">+</span>
    #             #     <span class="attribute" title="Attribute">min</span>:
    #             #     <span class="type" title="datetime">datetime</span>
    #             #     <code class="intermediate">0001-01-01 00:00:00</code>
    #             # </li>
    #             # """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">minute</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_tz_time.minute),
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">month</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_tz_time.month),
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">resolution</span>:
    #                 <span class="type" title="timedelta">timedelta</span>
    #                 <code class="intermediate">0:00:00.000001</code>
    #             </li>
    #             ""","""
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">second</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_tz_time.second),
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">tzinfo</span>:
    #                 <span class="type" title="timezone">timezone</span>
    #                 <code class="intermediate">UTC</code>
    #             </li>
    #             """,
    #             """
    #             <li>
    #                 <span class="access-modifier">+</span>
    #                 <span class="attribute" title="Attribute">year</span>:
    #                 <span class="type" title="int">int</span>
    #                 <code class="number">{0}</code>
    #             </li>
    #             """.format(now_tz_time.year),
    #
    #             # Object closing tags.
    #             """
    #                                 </div>
    #                             </ul>
    #                         <ul class="attribute-list"></ul>
    #                     </div>
    #                 <span class="braces">}</span>
    #             </div>
    #             """,
    #
    #             '<hr>',
    #         ],
    #         content_starts_after='sample_dt_datetime',
    #         content_ends_before='sample_dt_time',
    #     )

    @freeze_time(now_dt_time)
    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_dt_time_display(self, mocked_unique_generation):
        """Verify dumping a "datetime time" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_dt_time</span>
                    </span>:
                    <span class="type" title="time">time</span>
                """,
                """
                <code class="intermediate">{0}</code>
                """.format(now_dt_time.time()),
                """
                <span class="braces">{</span>
                <a
                    class="arrow-toggle collapsed"
                    title="[Ctrl+click] Expand all children"
                    data-toggle="collapse"
                    data-target=".data_9001"
                    data-dd-type="type"
                    data-object-depth="1"
                    aria-label="Close"
                    aria-expanded="false"
                >
                    <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                    <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                </a>
                <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                    <ul class="attribute-list">
                        <a
                            class="arrow-toggle show always-show"
                            title="[Ctrl+click] Expand all children"
                            data-target=".data_9001-attributes"
                            data-dd-type="attr"
                            aria-label="Open/Close"
                            aria-expanded=""
                        >
                            <span class="section_name">Attributes</span>
                                <span id="arrow-data_9001-attributes" class="arrow">
                            </span>
                        </a>
                        <div
                            class="li-wrapper collapse data_9001-attributes show"
                            data-unique-attributes="data_9001-attributes"
                        >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">fold</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">0</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">hour</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.hour),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">max</span>:
                    <span class="type" title="time">time</span>
                    <code class="intermediate">23:59:59.999999</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">microsecond</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.microsecond),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">min</span>:
                    <span class="type" title="time">time</span>
                    <code class="intermediate">{0}</code>
                </li>
                """.format(now_dt_time.time().min),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">minute</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.minute),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">resolution</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">0:00:00.000001</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">second</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_dt_time.second),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">tzinfo</span>:
                    <span class="type" title="null">null</span>
                    <code class="none">None</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_tz_datetime',
            content_ends_before='sample_tz_time',
        )

    @freeze_time(now_tz_time)
    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_tz_time_display(self, mocked_unique_generation):
        """Verify dumping a "timezone time" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_tz_time</span>
                    </span>:
                    <span class="type" title="time">time</span>
                """,
                """
                <code class="intermediate">{0}</code>
                """.format(now_tz_time.time()),
                """
                <span class="braces">{</span>
                <a
                    class="arrow-toggle collapsed"
                    title="[Ctrl+click] Expand all children"
                    data-toggle="collapse"
                    data-target=".data_9001"
                    data-dd-type="type"
                    data-object-depth="1"
                    aria-label="Close"
                    aria-expanded="false"
                >
                    <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                    <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                </a>
                <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                    <ul class="attribute-list">
                        <a
                            class="arrow-toggle show always-show"
                            title="[Ctrl+click] Expand all children"
                            data-target=".data_9001-attributes"
                            data-dd-type="attr"
                            aria-label="Open/Close"
                            aria-expanded=""
                        >
                            <span class="section_name">Attributes</span>
                                <span id="arrow-data_9001-attributes" class="arrow">
                            </span>
                        </a>
                        <div
                            class="li-wrapper collapse data_9001-attributes show"
                            data-unique-attributes="data_9001-attributes"
                        >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">fold</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">0</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">hour</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_tz_time.hour),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">max</span>:
                    <span class="type" title="time">time</span>
                    <code class="intermediate">23:59:59.999999</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">microsecond</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_tz_time.microsecond),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">min</span>:
                    <span class="type" title="time">time</span>
                    <code class="intermediate">{0}</code>
                </li>
                """.format(now_tz_time.time().min),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">minute</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_tz_time.minute),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">resolution</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">0:00:00.000001</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">second</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">{0}</code>
                </li>
                """.format(now_tz_time.second),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">tzinfo</span>:
                    <span class="type" title="null">null</span>
                    <code class="none">None</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_dt_time',
            content_ends_before='sample_dt_timedelta',
        )

    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_dt_timedelta_display(self, mocked_unique_generation):
        """Verify dumping a "datetime timedelta" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_dt_timedelta</span>
                    </span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">1 day, 0:00:00</code>
                    <span class="braces">{</span>
                    <a
                        class="arrow-toggle collapsed"
                        title="[Ctrl+click] Expand all children"
                        data-toggle="collapse"
                        data-target=".data_9001"
                        data-dd-type="type"
                        data-object-depth="1"
                        aria-label="Close"
                        aria-expanded="false"
                    >
                        <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                        <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                    </a>
                    <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                        <ul class="attribute-list">
                            <a
                                class="arrow-toggle show always-show"
                                title="[Ctrl+click] Expand all children"
                                data-target=".data_9001-attributes"
                                data-dd-type="attr"
                                aria-label="Open/Close"
                                aria-expanded=""
                            >
                                <span class="section_name">Attributes</span>
                                    <span id="arrow-data_9001-attributes" class="arrow">
                                </span>
                            </a>
                            <div
                                class="li-wrapper collapse data_9001-attributes show"
                                data-unique-attributes="data_9001-attributes"
                            >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">days</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">1</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">max</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">999999999 days, 23:59:59.999999</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">microseconds</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">0</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">min</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">-999999999 days, 0:00:00</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">resolution</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">0:00:00.000001</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">seconds</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">0</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_tz_time',
            content_ends_before='sample_tz_timedelta',
        )

    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_tz_timedelta_display(self, mocked_unique_generation):
        """Verify dumping a "timezone timedelta" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_tz_timedelta</span>
                    </span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">2 days, 0:00:00</code>
                    <span class="braces">{</span>
                    <a
                        class="arrow-toggle collapsed"
                        title="[Ctrl+click] Expand all children"
                        data-toggle="collapse"
                        data-target=".data_9001"
                        data-dd-type="type"
                        data-object-depth="1"
                        aria-label="Close"
                        aria-expanded="false"
                    >
                        <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                        <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                    </a>
                    <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                        <ul class="attribute-list">
                            <a
                                class="arrow-toggle show always-show"
                                title="[Ctrl+click] Expand all children"
                                data-target=".data_9001-attributes"
                                data-dd-type="attr"
                                aria-label="Open/Close"
                                aria-expanded=""
                            >
                                <span class="section_name">Attributes</span>
                                    <span id="arrow-data_9001-attributes" class="arrow">
                                </span>
                            </a>
                            <div
                                class="li-wrapper collapse data_9001-attributes show"
                                data-unique-attributes="data_9001-attributes"
                            >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">days</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">2</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">max</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">999999999 days, 23:59:59.999999</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">microseconds</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">0</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">min</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">-999999999 days, 0:00:00</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">resolution</span>:
                    <span class="type" title="timedelta">timedelta</span>
                    <code class="intermediate">0:00:00.000001</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">seconds</span>:
                    <span class="type" title="int">int</span>
                    <code class="number">0</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_dt_timedelta',
            content_ends_before='Python pathlib examples:',
        )

    @unittest.skipIf(not PYTZ_PRESENT, 'Pytz not present. Likely Django >= 4.0.')
    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_pytz_timezone_display(self, mocked_unique_generation):
        """Verify dumping a "pytz timezone" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_pytz_timezone</span>
                    </span>:
                    <span class="type" title="pytz_timezone">pytz_timezone</span>
                    <code class="intermediate">UTC</code>
                    <span class="braces">{</span>
                    <a
                        class="arrow-toggle collapsed"
                        title="[Ctrl+click] Expand all children"
                        data-toggle="collapse"
                        data-target=".data_9001"
                        data-dd-type="type"
                        data-object-depth="1"
                        aria-label="Close"
                        aria-expanded="false"
                    >
                        <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                        <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                    </a>
                    <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                        <ul class="attribute-list">
                            <a
                                class="arrow-toggle show always-show"
                                title="[Ctrl+click] Expand all children"
                                data-target=".data_9001-attributes"
                                data-dd-type="attr"
                                aria-label="Open/Close"
                                aria-expanded=""
                            >
                                <span class="section_name">Attributes</span>
                                    <span id="arrow-data_9001-attributes" class="arrow">
                                </span>
                            </a>
                            <div
                                class="li-wrapper collapse data_9001-attributes show"
                                data-unique-attributes="data_9001-attributes"
                            >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">zone</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'UTC'</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_tz_timedelta',
            content_ends_before='Python pathlib examples:',
        )

    @unittest.skipIf(not ZONEINFO_PRESENT, 'ZoneInfo not present. Likely Python < 3.9.')
    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_zoneinfo_timezone_display(self, mocked_unique_generation):
        """Verify dumping a "zoneinfo timezone" type has expected output."""

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_zoneinfo_timezone</span>
                    </span>:
                    <span class="type" title="ZoneInfo">ZoneInfo</span>
                    <code class="intermediate">UTC</code>
                    <span class="braces">{</span>
                    <a
                        class="arrow-toggle collapsed"
                        title="[Ctrl+click] Expand all children"
                        data-toggle="collapse"
                        data-target=".data_9001"
                        data-dd-type="type"
                        data-object-depth="1"
                        aria-label="Close"
                        aria-expanded="false"
                    >
                        <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                        <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                    </a>
                    <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                        <ul class="attribute-list">
                            <a
                                class="arrow-toggle show always-show"
                                title="[Ctrl+click] Expand all children"
                                data-target=".data_9001-attributes"
                                data-dd-type="attr"
                                aria-label="Open/Close"
                                aria-expanded=""
                            >
                                <span class="section_name">Attributes</span>
                                    <span id="arrow-data_9001-attributes" class="arrow">
                                </span>
                            </a>
                            <div
                                class="li-wrapper collapse data_9001-attributes show"
                                data-unique-attributes="data_9001-attributes"
                            >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">key</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'UTC'</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_tz_timedelta',
            content_ends_before='Python pathlib examples:',
        )

    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_pure_path_display(self, mocked_unique_generation):
        """Verify dumping a "pure path" type has expected output."""

        # Get project root dir.
        # Verify it is of either format "django-dump-die" or "django_dump_die".
        # Allows tests to handle regardless of how local user prefers local project naming.
        ROOT_DIR = os.path.basename(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
        try:
            self.assertEqual(ROOT_DIR, 'django-dump-die')
        except AssertionError:
            try:
                self.assertEqual(ROOT_DIR, 'django_dump_die')
            except AssertionError:
                err_msg = 'Failed to parse project root folder name. Recieved unexpected value of "{0}"'.format(
                    ROOT_DIR,
                )
                raise AssertionError(err_msg)

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                    <span class="dumped_name">sample_pure_path</span>
                </span>:
                <span class="type" title="PurePosixPath">PurePosixPath</span>
                """,
                """
                <code class="intermediate">{0}</code>
                """.format(project_path),
                """
                <span class="braces">{</span>
                <a
                    class="arrow-toggle collapsed"
                    title="[Ctrl+click] Expand all children"
                    data-toggle="collapse"
                    data-target=".data_9001"
                    data-dd-type="type"
                    data-object-depth="1"
                    aria-label="Close"
                    aria-expanded="false"
                >
                    <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                    <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                </a>
                <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                    <ul class="attribute-list">
                        <a
                            class="arrow-toggle show always-show"
                            title="[Ctrl+click] Expand all children"
                            data-target=".data_9001-attributes"
                            data-dd-type="attr"
                            aria-label="Open/Close"
                            aria-expanded=""
                        >
                            <span class="section_name">Attributes</span>
                            <span id="arrow-data_9001-attributes" class="arrow"></span>
                        </a>
                        <div
                            class="li-wrapper collapse data_9001-attributes show"
                            data-unique-attributes="data_9001-attributes"
                        >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">anchor</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'/'</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">drive</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">name</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'{0}'</code>
                </li>
                """.format(ROOT_DIR),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">parent</span>:
                    <span class="type" title="PurePosixPath">PurePosixPath</span>
                    <code class="intermediate">{0}</code>
                </li>
                """.format(os.path.dirname(project_path)),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">parents</span>:
                    <span class="type" title="_PathParents">_PathParents</span>
                    <code class="default"><PurePosixPath.parents></code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">parts</span>:
                    <span class="type" title="tuple">tuple</span>
                    <code class="default">{0}</code>
                </li>
                """.format(tuple(['/'] + project_path.split('/')[1:])),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">root</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'/'</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">stem</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'{0}'</code>
                </li>
                """.format(ROOT_DIR),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">suffix</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">suffixes</span>:
                    <span class="type" title="list">list</span>
                    <code class="default">[]</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='Python pathlib examples:',
            content_ends_before='sample_posix_path',
        )

    @patch('django_dump_die.templatetags.dump_die._generate_unique')
    def test_posix_path_display(self, mocked_unique_generation):
        """Verify dumping a "posix path" type has expected output."""

        # Get project root dir.
        # Verify it is of either format "django-dump-die" or "django_dump_die".
        # Allows tests to handle regardless of how local user prefers local project naming.
        ROOT_DIR = os.path.basename(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
        try:
            self.assertEqual(ROOT_DIR, 'django-dump-die')
        except AssertionError:
            try:
                self.assertEqual(ROOT_DIR, 'django_dump_die')
            except AssertionError:
                err_msg = 'Failed to parse project root folder name. Recieved unexpected value of "{0}"'.format(
                    ROOT_DIR,
                )
                raise AssertionError(err_msg)

        # Override default "unique" generation logic, for reproduce-able tests.
        # This generates enough uniques to guarantee mock does not raise errors.
        side_effects = []
        for index in range(5000):
            side_effects += [
                (f'data_9001', ''),
            ]
        mocked_unique_generation.side_effect = side_effects

        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',

                # Object opening tags.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_posix_path</span>
                    </span>:
                    <span class="type" title="PosixPath">PosixPath</span>
                """,
                """
                    <code class="intermediate">{0}</code>
                """.format(project_path),
                """
                    <span class="braces">{</span>
                    <a
                        class="arrow-toggle collapsed"
                        title="[Ctrl+click] Expand all children"
                        data-toggle="collapse"
                        data-target=".data_9001"
                        data-dd-type="type"
                        data-object-depth="1"
                        aria-label="Close"
                        aria-expanded="false"
                    >
                        <span class="unique" data-highlight-unique="data_9001">data_9001</span>
                        <span id="arrow-data_9001" class="arrow arrow-data_9001">▶</span>
                    </a>
                    <div class="dd-wrapper collapse data_9001 " data-unique="data_9001">
                        <ul class="attribute-list">
                            <a
                                class="arrow-toggle show always-show"
                                title="[Ctrl+click] Expand all children"
                                data-target=".data_9001-attributes"
                                data-dd-type="attr"
                                aria-label="Open/Close"
                                aria-expanded=""
                            >
                                <span class="section_name">Attributes</span>
                                    <span id="arrow-data_9001-attributes" class="arrow">
                                </span>
                            </a>
                            <div
                                class="li-wrapper collapse data_9001-attributes show"
                                data-unique-attributes="data_9001-attributes"
                            >
                """,

                # Object child elements.
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">anchor</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'/'</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">drive</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">name</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'{0}'</code>
                </li>
                """.format(ROOT_DIR),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">parent</span>:
                    <span class="type" title="PosixPath">PosixPath</span>
                    <code class="intermediate">{0}</code>
                </li>
                """.format(os.path.dirname(project_path)),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">parents</span>:
                    <span class="type" title="_PathParents">_PathParents</span>
                    <code class="default"><PosixPath.parents></code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">parts</span>:
                    <span class="type" title="tuple">tuple</span>
                    <code class="default">{0}</code>
                </li>
                """.format(tuple(['/'] + project_path.split('/')[1:])),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">root</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'/'</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">stem</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'{0}'</code>
                </li>
                """.format(ROOT_DIR),
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">suffix</span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </li>
                """,
                """
                <li>
                    <span class="access-modifier">+</span>
                    <span class="attribute" title="Attribute">suffixes</span>:
                    <span class="type" title="list">list</span>
                    <code class="default">[]</code>
                </li>
                """,

                # Object closing tags.
                """
                                    </div>
                                </ul>
                            <ul class="attribute-list"></ul>
                        </div>
                    <span class="braces">}</span>
                </div>
                """,

                '<hr>',
            ],
            content_starts_after='sample_pure_path',
            content_ends_before='done',
        )




from .dd_view import dd_view
from .example_views import (
    index,

    simple_type_example,
    intermediate_type_example,
    complex_type_example,
    function_type_example,
    class_type_example,
    full_category_example,

    numeric_example,
    datetime_example,
    django_model_example,
    iterable_group_example,
    system_path_example,
    full_purpose_example,

    django_request_response_cycle_example,
    edge_case_example,
)



"""
Project testing settings, so that tests can run project as if it was a proper Django application.
"""

# System Imports.
import sys


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    },
}

INSTALLED_APPS = (
    'django.contrib.sessions',
    'django.contrib.contenttypes',
    'django.contrib.staticfiles',
    'django.contrib.auth',
    'django_dump_die',
)

MIDDLEWARE = [
    'django_dump_die.middleware.DumpAndDieMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
]

ROOT_URLCONF = 'tests.urls'

USE_TZ = True

TIME_ZONE = 'UTC'

SECRET_KEY = 'test_secret_key'

TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'APP_DIRS': True,
}]


STATIC_URL = '/static/'


DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

# This should normally be False for tests, but since these tests are testing
# a tool used for development and debugging, it must be True.
DEBUG = True


# Suppress or show testcase debug printout, based on UnitTest execution method.
if 'pytest' in sys.modules:
    # Running Pytest env.
    # Pytest only shows console output on test failure, so we want it on.
    DJANGO_EXPANDED_TESTCASES_DEBUG_PRINT = True
else:
    # Running other testing env (mostly likely "django manage.py test").
    # manage.py shows all console output always, even on success.
    # So we want it off to avoid information overload and spam.
    DJANGO_EXPANDED_TESTCASES_DEBUG_PRINT = False



"""Core view for DumpDie library. All logic renders from this."""

# Third-Party Imports.
from django.conf import settings
from django.shortcuts import render


def dd_view(request, objects):
    """
    Return DumpDie view.
    :param request: Request object.
    :param objects: Set of objects to dump.
    """
    # Get toolbar options.
    include_util_toolbar = getattr(settings, 'DJANGO_DD_INCLUDE_UTILITY_TOOLBAR', True)
    attrs_enabled = getattr(settings, 'DJANGO_DD_INCLUDE_ATTRIBUTES', True)
    funcs_enabled = getattr(settings, 'DJANGO_DD_INCLUDE_FUNCTIONS', False)

    # Get user theme choices.
    force_light_theme = getattr(settings, 'DJANGO_DD_FORCE_LIGHT_THEME', False)
    force_dark_theme = getattr(settings, 'DJANGO_DD_FORCE_DARK_THEME', False)
    custom_color_theme = getattr(settings, 'DJANGO_DD_COLOR_SCHEME', None)
    multiline_function_docs = getattr(settings, 'DJANGO_DD_MULTILINE_FUNCTION_DOCS', False)

    # Validate chosen themes.
    if force_light_theme and force_dark_theme:
        raise ValueError("You can't force both light and dark themes.")

    # Render template.
    return render(request, 'django_dump_die/dd.html', {
        'objects': objects,
        'include_util_toolbar': include_util_toolbar,
        'attrs_enabled': attrs_enabled,
        'funcs_enabled': funcs_enabled,
        'force_light_theme': force_light_theme,
        'force_dark_theme': force_dark_theme,
        'custom_color_theme': custom_color_theme,
        'multiline_function_docs': multiline_function_docs,
    })



"""
Testing URL configuration for django-dump-die project.

Mocks being the "project settings root" urls.py file.
"""

# System Imports.
from django.urls import include, path


urlpatterns = [
    path('tests/', (include('django_dump_die.test_urls', namespace='django_dump_die_tests'))),
    path('', (include('django_dump_die.urls', namespace='django_dump_die'))),
]
